## Depreciated Variables ##
# FLOPS
flops_deprecated_vars = [
    'AERIN.MODARO',
    'ENGDIN.IGENEN',
    'ENGDIN.BOOST',
    'ENGDIN.EXTFAC',
    'ENGDIN.IXTRAP',
    'ENGDIN.NPCODE',
    'ENGDIN.PCODE',
    'ENGDIN.NOX',
    'OPTION.IOPT',
    'OPTION.IPOLP',
    'OPTION.NOISE'
]

# GASP
gasp_deprecated_vars = []
